/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

/*
begin_html
<p>
 See also:
 <ul>
 <li><a href="max_subi_ll.h.html">max_subi_ll.h</a></li>
 <li><a href="max_subi_ll.py.html">max_subi_ll.py</a><li>
 <li><a href="max_ll_util.h.html">max_ll_util.h</a></li>
 <li><a href="max_inter_ll.h.html">max_inter_ll.h</a></li>
 <li><a href="ublas_matrixutil.h.html">ublas_matrixutil.h</a></li>
 <li><a href="callbk_reportstream.h.html">callbk_reportstream.h</a></li>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
</p>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <string.h>
#include <iostream>
#include <fstream>

#include <string.h>
#include <set>
#include <map>
#include <vector>
#include <algorithm>

#include "max_subi_ll.h"
#include "max_inter_ll.h"
#include "max_ll_util.h"
#include "ublas_matrixutil.h"
#include "callbk_reportstream.h"

using std::cout;
using std::cerr;
using std::endl;
using std::set;
using std::map;
using std::pair;
using std::copy;

extern "C" MAXILL_API void __stdcall msl_multiplex_bounds(int Nseg, int Nsignal,
							  int **dwellCountses, int ***classeseses,
							  float ***durationseses, double ***ampseses,
							  int *out_dwellCounts, int *out_Nplex, int *out_Nstim)
{
  // First we need to know Nc[s] -- max number of classes in signal s
  typedef std::vector< set<int> > setlist;
  setlist classets(Nsignal);
  for (int seg=0; seg<Nseg; ++seg) {
    int *dwellCounts = dwellCountses[seg];
    int **classeses = classeseses[seg];
    for (int sig=0; sig<Nsignal; ++sig) {
      set<int>& classet = classets[sig];
      int *classes_end = classeses[sig]+dwellCounts[sig];
      for (int *cl=classeses[sig]; cl != classes_end; ++cl)
	classet.insert(*cl);
    }
  }
  std::vector<int> Nc;
  for (setlist::iterator csi=classets.begin(); csi != classets.end(); ++csi)
    Nc.push_back(csi->size());

  // If every combination of stimuli actually occurs, there will be
  //   Nc[1] * Nc[2] * ... stimclasses
  *out_Nstim = 1;
  for (int i=1; i<Nsignal; ++i)
    *out_Nstim *= Nc[i];
  
  // If every combination of stimclass and Markov-class occurs, there will be
  //   Nstim * Nc[0] plexiclasses
  *out_Nplex = *out_Nstim * Nc[0];
  
  // If none of the transitions on any signals coincide in time, there will be fewer than
  //   dwellCounts[0] + dwellCounts[1] + ... events in a segment
  for (int seg=0; seg<Nseg; ++seg) {
    out_dwellCounts[seg] = 0;
    for (int sig=0; sig<Nsignal; ++sig)
      out_dwellCounts[seg] += dwellCountses[seg][sig];
  }
}


typedef pair<int,int> plex;
typedef map<plex, int> plexmap;
typedef map<std::vector<double>, int> stimmap;

void msl_multiplex_segment(int Nsignal, int *dwellCounts, int **classeses, float **durationses,
			   double **ampses, int *out_dwellCount, int *out_classes, float *out_durations,
			   plexmap& plexcls_of, stimmap& stimcls_of)
{
  if ( ! Nsignal )
    return;
  // the key to stimmap
  std::vector<double> amps(Nsignal);
  // output
  int &Nd = *out_dwellCount;
  Nd = 0;
  // per-signal stuff
  std::vector<int> finger(Nsignal, 0);
  std::vector<int> clss(Nsignal);
  std::vector<float> remain(Nsignal);
  for (int i=0; i<Nsignal; ++i) {
    if ( ! dwellCounts[i] )
      return;
    clss[i] = classeses[i][0];
    remain[i] = durationses[i][0];
    if ( i ) { // the Markov class isn't part of stimclass
      if ( clss[i] >= 0 )
	amps[i] = ampses[i][ clss[i] ];
      else
	amps[i] = 0.0; // gap at beginning of stimulus: should complain and fail; instead silently defaulting to zero
    }
  }
  while ( 1 ) {
    // until any signal ends
    bool done = false;
    for (int i=0; i<Nsignal; ++i)
      if ( finger[i] >= dwellCounts[i] )
	done = true;
    if ( done )
      break;
    // find signal with shortest remaining
    float tm = remain[0];
    int changing = 0;
    for (int i=0; i<Nsignal; ++i) {
      if ( remain[i] < tm ) {
	tm = remain[i];
	changing = i;
      }
    }
    // look up stim- and plexi-classes
    int sc, pc;
    stimmap::iterator sti = stimcls_of.find(amps);
    if ( sti == stimcls_of.end() ) {
      sc = stimcls_of.size();
      stimcls_of[amps] = sc;
    } else {
      sc = stimcls_of[amps];
    }
    plex plx(clss[0], sc);
    plexmap::iterator pli = plexcls_of.find(plx);
    if ( pli == plexcls_of.end() ) {
      pc = plexcls_of.size();
      plexcls_of[plx] = pc;
    } else {
      pc = plexcls_of[plx];
    }
    // append the event
    out_classes[Nd] = pc;
    out_durations[Nd] = tm;
    ++Nd;
    // shorten all signals remain, and move the finger for any that are 0
    for (int i=0; i<Nsignal; ++i) {
      remain[i] -= tm;
      if ( remain[i] <= 0.0 ) {
	if ( (++finger[i]) < dwellCounts[i] ) {
	  remain[i] = durationses[i][ finger[i] ];
	  clss[i] = classeses[i][ finger[i] ];
	  if ( i )
	    if ( clss[i] >= 0 )
	      amps[i] = ampses[i][ clss[i] ];
	    // else: gap: keep the last event's amp
	}
      }
    }
  }
}

extern "C" MAXILL_API void __stdcall msl_multiplex(int Nseg, int Nsignal,
						   int **dwellCountses, int ***classeseses,
						   float ***durationseses, double ***ampseses,
						   double tdead_ms, int *out_dwellCounts,
						   int **out_classeses, float **out_durationses,
						   int *out_Nplex, int *out_plexiclasses,
						   int *out_Nstim, double *out_stimclasses)
{
  // We use the same plexi- and stim-classes for all segments, numbered in order of appearance.
  // We'll collect the indices in these maps
  plexmap plexcls_of;
  stimmap stimcls_of;
  
  for (int seg=0; seg<Nseg; ++seg)
    msl_multiplex_segment(Nsignal, dwellCountses[seg], classeseses[seg], durationseses[seg],
			  ampseses[seg], out_dwellCounts+seg, out_classeses[seg],
			  out_durationses[seg], plexcls_of, stimcls_of);

  // apply dead time
  mil_prep_segments(Nseg, out_dwellCounts, out_classeses, out_durationses, tdead_ms);

  // and reverse s- and p-classes out
  *out_Nplex = plexcls_of.size();
  for (plexmap::iterator plexi = plexcls_of.begin(); plexi != plexcls_of.end(); ++plexi) {
    int i = plexi->second * 2;
    out_plexiclasses[i] = plexi->first.first;
    out_plexiclasses[i+1] = plexi->first.second;
  }
  *out_Nstim = 0;
  for (stimmap::iterator stimi = stimcls_of.begin(); stimi != stimcls_of.end(); ++stimi) {
    int i = stimi->second * Nsignal;
    copy(stimi->first.begin(), stimi->first.end(), out_stimclasses + i);
	++(*out_Nstim);
  }
}



extern "C" MAXILL_API int __stdcall subi_ll_arr(
	int Ns, int *clazz, double *P0_in,  double *K0_in, double *K1_in,
	                                    int *Ligand_in, int *Voltage_in,
	double tdead_sec, int Nseg, int *dwellCounts, int **classeses, float **durationses,
	int Nsig, int Nplex, int *plexiclasses, int Nstim, double *stimclasses,
	double *ll, eigen_func custom_eigen)
{
  std::vector<double*> K0, K1;
  std::vector<int*> Lig, Vol;
  for (int i=0; i<Ns; ++i) {
    K0.push_back( K0_in + i*Ns );
    K1.push_back( K1_in + i*Ns );
    Lig.push_back( Ligand_in + i*Ns );
    Vol.push_back( Voltage_in + i*Ns );
  }
  std::vector<double> segLL(Nseg);

  return subi_ll(Ns, clazz, P0_in, &(K0[0]), &(K1[0]), &(Lig[0]), &(Vol[0]),
		 tdead_sec, Nseg, dwellCounts, classeses, durationses,
		 Nsig, Nplex, plexiclasses, Nstim, stimclasses, ll, custom_eigen, &(segLL[0]), NULL, NULL, NULL);
}


extern "C" MAXILL_API int __stdcall subi_ll(
	int Ns, int *clazz, double *P0_in,  double **K0_in, double **K1_in,
	int **Ligand_in, int **Voltage_in,
	double tdead_sec, int Nseg, int *dwellCounts, int **classeses, float **durationses,
	int Nsig, int Nplex, int *plexiclasses, int Nstim, double *stimclasses,
	double *ll, eigen_func custom_eigen, double *segLL, callbk_reportfun report_cb, void *report_data, int *stop_flag)
{
  callbk_reportstream output(report_cb, report_data);

  // find max class in model and data -- make sure it's possible
  int maxcls = clazz[0];
  for (int i=1; i<Ns; ++i)
    if ( clazz[i] > maxcls )
      maxcls = clazz[i];
  
  int maxcls_data = -1;
  for (int i=0; i<Nplex; ++i) {
    if ( plexiclasses[2*i] > maxcls_data )
      maxcls_data = plexiclasses[2*i];
  }
  if ( maxcls_data > maxcls ) {
    output << "MSL: Not enough classes in the model." << endl;
    return -1;
  }
  
  // list the states that are in each class, and the states which are not:
  std::vector< std::vector<int> > clazz_states(maxcls+1);
  std::vector< std::vector<int> > clazz_oppos(maxcls+1);
  for (int i=0; i<Ns; ++i) {
    clazz_states[clazz[i]].push_back(i);
    for (int j=0; j<Ns; ++j)
      if ( clazz[j] != clazz[i] )
	clazz_oppos[clazz[i]].push_back(j);
  }

  // copy inputs into ublas objects
  matrix<double> K0;         matrix_from_arrays(K0, K0_in, Ns, Ns);
  matrix<double> K1;         matrix_from_arrays(K1, K1_in, Ns, Ns);
  matrix<int>    Ligand;     matrix_from_arrays(Ligand, Ligand_in, Ns, Ns);
  matrix<int>    Voltage;    matrix_from_arrays(Voltage, Voltage_in, Ns, Ns);

  // set up P0, eQ, eigen-decomposition of submatrices eQaa, and A(tdead)
  // and temporary matrices for intermediate calculations
  
  // normalize fixed P0 (if any) to sum to 1
  matrix<double> P0(1, Ns);
  if ( P0_in ) 
  {
    matrix_from_array(P0, P0_in, 1, Ns);
    double sum = 0.0;
    for (int i=0; i<Ns; ++i)
      sum += P0_in[i];
    for (int i=0; i<Ns; ++i)
      P0(0,i) = (sum > 1e-8 ) ? (P0(0,i) / sum) : (1.0 / Ns);
  }
  int            Na;

  // build dead-time-corrected Q matrices, different in each stimclass
  std::vector< matrix<double> > QQ(Nstim), eQQ(Nstim);
  for (int sc=0; sc<Nstim; ++sc) {
    BuildQ(Ns, QQ[sc], K0, K1, Ligand, Voltage, stimclasses + sc*Nsig);
    QtoQe(QQ[sc], clazz, tdead_sec, eQQ[sc]);
  }
  
  // we eigendecompose each eQ for quick exponentiation in the inner loop: (gap A matrix)
  // indices are [stimclass][(vector or matrix)]
  std::vector< std::vector<double> > eigvals_eq(Nstim); // (maxcls+1);
  std::vector< matrix<double> > eigvecs_eq(Nstim);
  std::vector< matrix<double> > eiginvs_eq(Nstim);
  vector<double> eig_imag(Ns);
  for (int sc=0; sc<Nstim; ++sc) {
    if ( stop_flag && *stop_flag ) return -33;
    eigvals_eq[sc].resize(Ns);
    eigvecs_eq[sc].resize(Ns, Ns);
    eiginvs_eq[sc].resize(Ns, Ns);
    bool err_eigen;
    if ( custom_eigen )
      err_eigen = custom_eigen(Ns, & eQQ[sc](0,0), & eigvals_eq[sc][0], & eig_imag[0],
			       & eigvecs_eq[sc](0,0), & eiginvs_eq[sc](0,0));
    else
      err_eigen = ! eigen_vals_vecs(eQQ[sc], &(eigvals_eq[sc][0]), &(eig_imag[0]), eigvecs_eq[sc], eiginvs_eq[sc], output);
    if ( err_eigen ) {
      output << "No convergence in eigen calculations" << endl;
      return -2;
    }
    for (int i=0; i<Ns; ++i) {
      if ( fabs(eig_imag[i]) > 1e-5 ) {
	output << "Complex eigenvalues found: " << eig_imag[i] << endl;
	return -2;
      }
      if ( eigvals_eq[sc][i] > 1e-5 ) {
	output << "Positive eigenvalues found: " << eigvals_eq[sc][i] << endl;
	return -2;
      }
    }
  }
  // we eigendecompose each submatrix eQaa for quick exponentiation in the inner loop:
  // indices are [stimclass][model class][(vector or matrix)]
  std::vector< std::vector< std::vector<double> > > eigvals_aa(Nstim); // (maxcls+1);
  std::vector< std::vector< matrix<double> > > eigvecs_aa(Nstim);
  std::vector< std::vector< matrix<double> > > eiginvs_aa(Nstim);
  for (int sc=0; sc<Nstim; ++sc) {
    if ( stop_flag && *stop_flag ) return -33;
    eigvals_aa[sc].resize(maxcls+1);
    eigvecs_aa[sc].resize(maxcls+1);
    eiginvs_aa[sc].resize(maxcls+1);
    for (int a=0; a<=maxcls; ++a) {
      Na = clazz_states[a].size();
      if ( Na ) {
	int *aa = &(clazz_states[a][0]);
	matrix<double> eQaa;   matrix_to_partition(eQQ[sc], Na, Na, aa, aa, eQaa);
	eigvals_aa[sc][a].resize(Na);
	eigvecs_aa[sc][a].resize(Na, Na);
	eiginvs_aa[sc][a].resize(Na, Na);
	if ( ! eigen_vals_vecs(eQaa, &(eigvals_aa[sc][a][0]), &(eig_imag[0]), eigvecs_aa[sc][a], eiginvs_aa[sc][a], output) ) {
	  output << "No convergence in eigen calculations" << endl;
	  return -2;
	}
	for (int i=0; i<Na; ++i) {
	  if ( fabs(eig_imag[i]) > 1e-5 ) {
	    output << "Complex eigenvalues found: " << eig_imag[i] << endl;
	    return -2;
	  }
	  if ( eigvals_aa[sc][a][i] > 1e-5 ) {
	    output << "Positive eigenvalues found: " << eigvals_aa[sc][a][i] << endl;
	    return -2;
	  }
	}
      }
    }
  }
  // Then for efficiency, allocate space for expm_eigen temps
  std::vector< matrix<double> > tmp_nxn(Ns+1), tmp_nxn_diag(Ns+1);
  for (int i=1; i<=Ns; ++i) {
    tmp_nxn[i].resize(i, i);
    tmp_nxn_diag[i].resize(i, i);
    for (int j=0; j<i; ++j)
      for (int k=0; k<i; ++k)
	tmp_nxn_diag[i](j,k) = 0.0;
  }
  // and a place for the final exponentiated matrices:
  std::vector< matrix<double> > eQ_aa(maxcls+1);
  for (int a=0; a<=maxcls; ++a) {
    Na = clazz_states[a].size();
    if ( Na )
      eQ_aa[a].resize(Na, Na);
  }
  // which get copied out to the full Ns x Ns
  matrix<double> eQ_aa_full(Ns, Ns);
  
  // each stimclass has Atd = exp(Q * tdead)
  std::vector< matrix<double> > AAtd(Nstim);
  for (int sc=0; sc<Nstim; ++sc) {
    if ( stop_flag && *stop_flag ) return -33;
	AAtd[sc].resize(Ns, Ns);
    expm(QQ[sc], tdead_sec, AAtd[sc]);
  }
  // for a given destination class a, we zero out columns belonging to other classes
  // (probability of ending outside a is 0, given the data)
  std::vector< std::vector< matrix<double> > > AAtd_to(Nstim); // [sc][a]
  for (int sc=0; sc<Nstim; ++sc) {
    matrix<double>& Atd = AAtd[sc];
    AAtd_to[sc].resize(maxcls+1);
    for (int a=0; a<=maxcls; ++a) {
      matrix<double>& Atd_a = AAtd_to[sc][a];
      std::vector<int>::iterator other = clazz_oppos[a].begin();
      std::vector<int>::iterator other_end = clazz_oppos[a].end();
      Atd_a.resize(Ns, Ns);
      for (int j=0; j<Ns; ++j) {
	if ( (other != other_end) && (*other == j) ) {
	  ++other;
	  for (int i=0; i<Ns; ++i)
	    Atd_a(i,j) = 0.0;
	}
	else {
	  for (int i=0; i<Ns; ++i)
	    Atd_a(i,j) = Atd(i,j);
	}
      }
    }
  }
  // scratch space for custom final Atd to "any class but a"
  matrix<double> Atd_final(Ns, Ns);

  // alternating temporary alpha rows per class to avoid temporaries alpha_a <- alpha_a * At
  vector<double> alpha_pre(Ns), alpha_post(Ns);

  // scale accumulator
  *ll = 0.0;
  
  // loop over segments, summing ll
  for (int iseg = 0; iseg<Nseg; ++iseg) {
    if ( stop_flag && *stop_flag ) return -33;
    int Nd = dwellCounts[iseg];
    int *classes = classeses[iseg];
    float *durations = durationses[iseg];
    int pc = classes[0];
    int a = plexiclasses[2*pc];
    int sc = plexiclasses[2*pc+1];
    if ( a >= 0 )
      Na = clazz_states[a].size();
    else // gap
      Na = Ns;
    int b = a;
    int Nb = Na;
    double &seg_ll = segLL[iseg];
    seg_ll = 0.0;
    
    // If they didn't give fixed P0, start at equilibrium for the starting stimulus
    if ( ! P0_in ) {
      QtoPe(eQQ[sc], P0);
    }
    
    // P0 are the entry probabilities, only in states of class a
    if ( a >= 0 ) { 
      for (int i=0; i<Ns; ++i) {
	alpha_pre(i) = (clazz[i] == a) ? P0(0, i) : 0.0;
      }
    } else { // gap: all states possible
      for (int i=0; i<Ns; ++i) {
	alpha_pre(i) = P0(0, i);
      }
    }
    // their contribution to ll is log( sum( alpha_pre ) )
    // and alpha_pre is normalized to sum to 1
    seg_ll += alpha_logscale(alpha_pre);
    
    // loop over all but the final dwell
    for (int d=1; d<Nd; ++d) {
      a = b;                      // next dwell, in class a
      Na = Nb;
      
      // check the user stop flag every so often
      if ( (! (d%512)) && stop_flag && *stop_flag ) return -33;

      /*
	begin_html
       in original MIL, scaling is done separately for the dwell and subsequent transition,
       and all alphas are kept for the gradient calculation.  Here we accumulate and scale both together.<br>
       note that dur has already had tdead subtracted<br>
       alpha <- alpha \(* e^{Q_{aa} * dur} * A_{td}\)[b]
        end_html
      */
      
      if ( a >= 0 ) {
	expm_eig(&(eigvals_aa[sc][a][0]), eigvecs_aa[sc][a], eiginvs_aa[sc][a], durations[d-1],
		 tmp_nxn_diag[Na], tmp_nxn[Na], eQ_aa[a]);
	eQ_aa_full.clear();
	matrix_from_partition(eQ_aa_full, Na, Na, &(clazz_states[a][0]), &(clazz_states[a][0]), eQ_aa[a]);
      } else { // gap: full exp(eQ)
	expm_eig(&(eigvals_eq[sc][0]), eigvecs_eq[sc], eiginvs_eq[sc], durations[d-1],
		 tmp_nxn_diag[Ns], tmp_nxn[Ns], eQ_aa_full);
      }
      noalias(alpha_post) = prod(alpha_pre, eQ_aa_full);
      // seg_ll += alpha_logscale(alpha_post);

      // switch sc after the dwell, before the transition
      pc = classes[d];
      b = plexiclasses[2*pc];    // going to class b
      sc = plexiclasses[2*pc+1];
      if ( b >= 0 )
	Nb = clazz_states[b].size();
      else // into gap
	Nb = Ns;
      
      if ( b >= 0 )
	noalias(alpha_pre) = prod(alpha_post, AAtd_to[sc][b]);
      else // into gap: all states possible
	noalias(alpha_pre) = prod(alpha_post, AAtd[sc]);
      seg_ll += alpha_logscale(alpha_pre);
    }
    // final dwell, in class a
    a = b;
    Na = Nb;
    // dwelling, same as above
    if ( a >= 0 ) {
      expm_eig(&(eigvals_aa[sc][a][0]), eigvecs_aa[sc][a], eiginvs_aa[sc][a], durations[Nd-1],
	       tmp_nxn_diag[Na], tmp_nxn[Na], eQ_aa[a]);
      eQ_aa_full.clear();
      matrix_from_partition(eQ_aa_full, Na, Na, &(clazz_states[a][0]), &(clazz_states[a][0]), eQ_aa[a]);
    }
    else {
      expm_eig(&(eigvals_eq[sc][0]), eigvecs_eq[sc], eiginvs_eq[sc], durations[Nd-1],
	       tmp_nxn_diag[Ns], tmp_nxn[Ns], eQ_aa_full);
    }
    noalias(alpha_post) = prod(alpha_pre, eQ_aa_full);
    // seg_ll += alpha_logscale(alpha_post);
    
    // and exiting to any class but a
    matrix<double>& Atd = AAtd[sc];
    if ( a >= 0 ) {
      Atd_final.clear();
      std::vector<int>::iterator other = clazz_oppos[a].begin();
      std::vector<int>::iterator other_end = clazz_oppos[a].end();
      for (int j=0; j<Ns; ++j) {
	if ( (other != other_end) && (*other == j) ) {
	  ++other;
	  for (int i=0; i<Ns; ++i)
	    Atd_final(i,j) = Atd(i,j);
	}
      }
    } else { // gap: no blanks
      Atd_final = Atd;
    }
    noalias(alpha_pre) = prod(alpha_post, Atd_final);
    seg_ll += alpha_logscale(alpha_pre);
    
    // now the one segment ll is done, add it to total ll
    *ll += seg_ll;
  }

  return 0; // success
}







